# Copyright (c) 2025 JP Hutchins
# SPDX-License-Identifier: MIT

"""Binary expressions for arithmetic, logic, and comparison operations.

## Syntax

It's recommended to employ static type analysis tools like `mypy` to provide
realtime feedback on the correctness of your expressions as you work.

### Mapping Context Variables <-> `Var`s

The rule is that the `str` name of the `Var`, it's `name` attribute, must match
the corresponding field name in the context object. For example, this will work:
>>> from typing import NamedTuple
>>> class Context(NamedTuple):
... 	my_name: str
>>> any_name_works_here = Var[int, Context]("my_name")
>>> any_name_works_here.to_string()
'my_name'
>>> any_name_works_here.to_string(Context(my_name='jp'))
'my_name:jp'

But this will not, when it's evaluated:
>>> wrong_name = Var[int, Context]("wrong_name")
>>> wrong_name.to_string()
'wrong_name'
>>> wrong_name.to_string(Context(my_name='jp'))
Traceback (most recent call last):
	...
mahonia.types.EvalError: Variable 'wrong_name' not found in context

This occurs because when `wrong_name` is evaluated, it looks for its' value at
the `wrong_name` attribute of the `Context` object, which does not exist.

Note that `Var`s are tied to their context type, so you cannot use a `Var` from
one context in another context and will get a static type error if you try.

### Type Coercion

Coercible types (like `int`, `float`, `bool`) can appear on either side of
a binary operator with a Mahonia `Expr`. The coercible type will be wrapped
in an unnamed `Const`.

>>> X = Const("X", 41)
>>> X + 1
Add(left=Const(name='X', value=41), right=Const(name=None, value=1))

>>> 1 + X
Add(left=Const(name=None, value=1), right=Const(name='X', value=41))

This works because `BinaryOperationOverloads` implements both the standard
operators (`__add__`, etc.) and their right-associative counterparts
(`__radd__`, etc.). When Python's `int.__add__` doesn't know how to handle
a Mahonia `Expr`, it falls back to the Expr's `__radd__` method.

Comparison operators also work with literals on the left, but Python reflects
them to the swapped operator (e.g., `5 < x` becomes `x > 5`):

>>> from typing import Any
>>> x = Var[int, Any]("x")
>>> 5 < x
Gt(left=Var(name='x'), right=Const(name=None, value=5))

For `==` and `!=`, static type checkers incorrectly infer `bool` because they
don't know that `int.__eq__` returns `NotImplemented` for unknown types. At
runtime, Python correctly falls back to the Expr's `__eq__` method. To avoid
type checker errors, prefer placing the Expr on the left side: `x == 5`.

### Examples

For more exhaustive examples, see the tests.

>>> from typing import NamedTuple, assert_type
...
>>> class Ctx(NamedTuple):
... 	x: int
... 	y: int
...
>>> x = Var[int, Ctx]("x")
>>> y = Var[int, Ctx]("y")
>>> MAX = Const("Max", 10)
>>> x
Var(name='x')
>>> MAX
Const(name='Max', value=10)
>>> # Create an expression that compares a Var and a Const
>>> x < MAX
Lt(left=Var(name='x'), right=Const(name='Max', value=10))
>>> # Serialize
>>> (x < MAX).to_string()
'(x < Max:10)'
>>> # Evaluate
>>> (x < MAX).unwrap(Ctx(x=5, y=10))
True
>>> # Serialize the evaluation
>>> (x < MAX).to_string(Ctx(x=5, y=10))
'(x:5 < Max:10 -> True)'
>>> # Assign an expression to a variable
>>> sum_expr = x + y
>>> sum_expr
Add(left=Var(name='x'), right=Var(name='y'))
>>> # Serialize the expression to a string
>>> sum_expr.to_string()
'(x + y)'
>>> # Evaluate the expression with concrete values
>>> sum_result = sum_expr(Ctx(x=1, y=2))
>>> sum_result
Const(name=None, value=3)
>>> # Serialize the evaluated expression to a string
>>> sum_expr.to_string(Ctx(x=1, y=2))
'(x:1 + y:2 -> 3)'
>>> # Access the value of the result
>>> sum_result.value
3
>>> # Shorthand for getting the value
>>> sum_expr.unwrap(Ctx(x=1, y=2))
3
>>> # Bind an expression to a context
>>> summed = sum_expr.bind(Ctx(x=1, y=2))
>>> str(summed)
'(x:1 + y:2 -> 3)'
>>> summed.unwrap()
3
>>> # Combine expressions
>>> zero = x + y - x - y
>>> zero.to_string()
'(((x + y) - x) - y)'
>>> zero.to_string(Ctx(x=1, y=2))
'(((x:1 + y:2 -> 3) - x:1 -> 2) - y:2 -> 0)'
>>> zero.unwrap(Ctx(x=1, y=2))
0
>>> # Compose expressions
>>> four = sum_expr + x
>>> four.to_string()
'((x + y) + x)'
>>> four.to_string(Ctx(x=1, y=2))
'((x:1 + y:2 -> 3) + x:1 -> 4)'
>>> four.unwrap(Ctx(x=1, y=2))
4
>>> # Create a predicate
>>> is_valid = (x > 0) & (y < MAX)
>>> is_valid.to_string()
'((x > 0) & (y < Max:10))'
>>> is_valid.to_string(Ctx(x=1, y=2))
'((x:1 > 0 -> True) & (y:2 < Max:10 -> True) -> True)'
>>> is_valid.unwrap(Ctx(x=1, y=2))
True
>>> # Define the predicate
>>> cate = Predicate("x is pos, y less than Max", is_valid)
>>> cate.to_string()
'x is pos, y less than Max: ((x > 0) & (y < Max:10))'
>>> cate.to_string(Ctx(x=1, y=2))
'x is pos, y less than Max: True ((x:1 > 0 -> True) & (y:2 < Max:10 -> True) -> True)'
>>> cate.unwrap(Ctx(x=1, y=2))
True
>>> # Other abstractions similar to Predicate include
>>> # Approximately and PlusMinus. PlusMinus is a kind
>>> # of Const that implements the ConstToleranceProtocol.
>>> # These can be use as reference for custom convenience
>>> # abstractions.
>>> class Ctx(NamedTuple):
... 	voltage: float
>>> voltage = Var[float, Ctx]("voltage")
>>> voltage_check = Approximately(voltage, PlusMinus("V", 5.0, 0.1))
>>> voltage_check.to_string(Ctx(voltage=5.05))
'(voltage:5.05 ≈ V:5.0 ± 0.1 -> True)'
"""

import operator
from dataclasses import dataclass
from typing import (
	TYPE_CHECKING,
	Any,
	Callable,
	ClassVar,
	Final,
	Generic,
	Protocol,
	overload,
	runtime_checkable,
)

from mahonia.types import ContextProtocol as ContextProtocol
from mahonia.types import EvalError as EvalError
from mahonia.types import MergeContextProtocol as MergeContextProtocol
from mahonia.types import R as R
from mahonia.types import R_Eval as R_Eval
from mahonia.types import S as S
from mahonia.types import S_contra as S_contra
from mahonia.types import SizedIterable as SizedIterable
from mahonia.types import Ss as Ss
from mahonia.types import SupportsArithmetic as SupportsArithmetic
from mahonia.types import SupportsComparison as SupportsComparison
from mahonia.types import SupportsEquality as SupportsEquality
from mahonia.types import SupportsLogic as SupportsLogic
from mahonia.types import T as T
from mahonia.types import T_co as T_co
from mahonia.types import TSupportsArithmetic as TSupportsArithmetic
from mahonia.types import TSupportsComparison as TSupportsComparison
from mahonia.types import TSupportsEquality as TSupportsEquality
from mahonia.types import TSupportsLogic as TSupportsLogic
from mahonia.types import U as U

if TYPE_CHECKING:
	from mahonia.match import Match as Match
	from mahonia.match import MatchExpr as MatchExpr


@dataclass(frozen=True)
class Failure:
	"""Accumulated errors from the Result (error-accumulating) context.

	>>> f = Failure((ValueError("bad x"), TypeError("bad y")))
	>>> len(f.exceptions)
	2
	>>> str(f.exceptions[0])
	'bad x'
	>>> str(f.exceptions[1])
	'bad y'
	"""

	exceptions: tuple[Exception, ...]


@runtime_checkable
class Eval(Protocol[R_Eval, S_contra]):
	"""Protocol for objects that can evaluate themselves with a context.

	This protocol defines the interface for expression evaluation. All Mahonia
	expressions implement this protocol. R_Eval is the result type returned by eval().

	>>> from typing import NamedTuple
	>>> class Ctx(NamedTuple):
	... 	x: int
	>>> x = Var[int, Ctx]("x")
	>>> isinstance(x, Eval)
	True
	>>> result = x.eval(Ctx(x=42))
	>>> result.value
	42
	"""

	def eval(self, ctx: S_contra) -> "Const[R_Eval]": ...


@runtime_checkable
class ToString(Protocol[S_contra]):
	"""Protocol for objects that can serialize themselves to strings.

	This protocol defines the interface for string serialization, with optional
	context for showing evaluated values.

	>>> from typing import NamedTuple
	>>> class Ctx(NamedTuple):
	... 	x: int
	>>> x = Var[int, Ctx]("x")
	>>> isinstance(x, ToString)
	True
	>>> x.to_string()
	'x'
	>>> x.to_string(Ctx(x=42))
	'x:42'
	"""

	def to_string(self, ctx: S_contra | None = None) -> str: ...


@runtime_checkable
class Expr(Protocol[T, S, R]):
	"""Base class for all Mahonia expressions.

	Provides core evaluation, serialization, and binding functionality.
	Users typically work with concrete expression types like Var, Const, Add, etc.

	Type parameters:
	- T: The operand type (for binary operations, the type of left/right)
	- S: The context type (bound by ContextProtocol)
	- R: The result type from eval() and unwrap(). Defaults to T for backward compat.

	>>> from typing import NamedTuple
	>>> class Ctx(NamedTuple):
	... 	x: int
	>>> x = Var[int, Ctx]("x")
	>>> isinstance(x, Expr)
	True
	>>> x.unwrap(Ctx(x=10))
	10
	>>> bound = x.bind(Ctx(x=10))
	>>> bound.unwrap()
	10
	"""

	def eval(self, ctx: S) -> "Const[R]": ...

	def to_string(self, ctx: S | None = None) -> str: ...

	def __call__(self, ctx: S) -> "Const[R]":
		return self.eval(ctx)

	def to_func(self) -> "Func[R, S]":
		return Func(extract_vars((), self), self)

	def map(self, container: "Expr[SizedIterable[Any], Any, Any]") -> "MapExpr[Any, R, Any]":
		"""Apply this expression as a function to each element in a container.

		This is shorthand for: self.to_func() -> MapExpr(func, container)

		>>> from typing import NamedTuple
		>>> class ElemCtx(NamedTuple):
		... 	x: int
		>>> class ContainerCtx(NamedTuple):
		... 	numbers: list[int]
		>>> x = Var[int, ElemCtx]("x")
		>>> numbers = Var[SizedIterable[int], ContainerCtx]("numbers")
		>>> mapped = (x * 2).map(numbers)
		>>> mapped.to_string()
		'(map x -> (x * 2) numbers)'
		"""
		func = self.to_func()
		return MapExpr(func, container)

	def unwrap(self, ctx: S) -> R:
		return self.eval(ctx).value

	def bind(self, ctx: S) -> "BoundExpr[T, S, R]":
		return BoundExpr(self, ctx)

	def partial(self, ctx: Any) -> "Expr[T, Any, R]":
		"""Partially apply a context, resolving variables that exist in it.

		Variables whose names exist as attributes in ctx are replaced with Const.
		Variables not found in ctx remain as Var. The expression structure is
		preserved (no eager evaluation of operations).

		>>> from typing import NamedTuple, Any
		>>> class XCtx(NamedTuple):
		... 	x: int
		>>> class YCtx(NamedTuple):
		... 	y: int
		>>> x = Var[int, Any]("x")
		>>> y = Var[int, Any]("y")
		>>> expr = x + y
		>>> partial_expr = expr.partial(XCtx(x=5))
		>>> partial_expr.to_string()
		'(x:5 + y)'
		>>> partial_expr.unwrap(YCtx(y=10))
		15
		>>> partial_expr.to_string(YCtx(y=10))
		'(x:5 + y:10 -> 15)'
		"""
		...


@dataclass(frozen=True, eq=False, slots=True)
class Func(Generic[T, S]):  # type: ignore[misc]
	args: tuple[Expr[Any, S, Any], ...]
	expr: Expr[Any, S, T]

	def to_string(self, ctx: S | None = None) -> str:
		if ctx is None:
			if len(self.args) == 0:
				return f"() -> {self.expr.to_string()}"
			elif len(self.args) == 1:
				return f"{self.args[0].to_string()} -> {self.expr.to_string()}"
			else:
				args_str = ", ".join(arg.to_string() for arg in self.args)
				return f"({args_str}) -> {self.expr.to_string()}"
		else:
			if len(self.args) == 0:
				return f"() -> {self.expr.to_string()} -> {self.expr.unwrap(ctx)}"
			elif len(self.args) == 1:
				return f"{self.args[0].unwrap(ctx)} -> {self.expr.to_string()} -> {self.expr.unwrap(ctx)}"
			else:
				args_str = ", ".join(f"{arg.unwrap(ctx)}" for arg in self.args)
				return f"({args_str}) -> {self.expr.to_string()} -> {self.expr.unwrap(ctx)}"


@runtime_checkable
class BoolExpr(Expr[T, S, bool], Protocol[T, S]):
	"""Base class for boolean expressions that support logical operations.

	Extends Expr with support for logical operators like & (and), | (or), and ~ (not).
	T is the operand type, S is the context type, and the result is always bool.

	>>> from typing import NamedTuple
	>>> class Ctx(NamedTuple):
	... 	x: int
	>>> x = Var[int, Ctx]("x")
	>>> bool_expr = x > 5
	>>> isinstance(bool_expr, BoolExpr)
	True
	>>> bool_expr.unwrap(Ctx(x=10))
	True
	>>> combined = bool_expr & (x < 20)
	>>> combined.unwrap(Ctx(x=10))
	True
	"""

	def eval(self, ctx: S) -> "Const[bool]": ...


class UnaryOperationOverloads(Expr[bool, S, bool]):
	def __invert__(self) -> "Not[S]":
		return Not(self)


class BooleanBinaryOperationOverloads(Generic[T, S]):
	@overload
	def __and__(  # type: ignore[misc]
		self: "ResultBase[Any, S]", other: "Expr[Any, S, Any]"
	) -> "Result[bool, S, bool]": ...

	@overload
	def __and__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[Any, S]"
	) -> "Result[bool, S, bool]": ...

	@overload
	def __and__(self, other: "Expr[Any, S, Any]") -> "And[bool, S]": ...

	def __and__(self, other: "Expr[Any, S, Any]") -> "And[bool, S] | Result[bool, S, bool]":
		rhs = other if isinstance(other, Expr) else Const(None, other)  # pyright: ignore[reportUnnecessaryIsInstance]
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				And,
				self if isinstance(self, ResultBase) else Pure(self),  # type: ignore[arg-type]  # pyright: ignore[reportUnknownArgumentType]
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
			)
		return And(self, rhs)  # type: ignore[arg-type]

	def __rand__(self, other: bool) -> "And[bool, S]":
		return And(Const(None, other), self)  # type: ignore[arg-type]

	@overload
	def __or__(  # type: ignore[misc]
		self: "ResultBase[Any, S]", other: "Expr[Any, S, Any]"
	) -> "Result[bool, S, bool]": ...

	@overload
	def __or__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[Any, S]"
	) -> "Result[bool, S, bool]": ...

	@overload
	def __or__(self, other: "Expr[Any, S, Any]") -> "Or[bool, S]": ...

	def __or__(self, other: "Expr[Any, S, Any]") -> "Or[bool, S] | Result[bool, S, bool]":
		rhs = other if isinstance(other, Expr) else Const(None, other)  # pyright: ignore[reportUnnecessaryIsInstance]
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Or,
				self if isinstance(self, ResultBase) else Pure(self),  # type: ignore[arg-type]  # pyright: ignore[reportUnknownArgumentType]
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
			)
		return Or(self, rhs)  # type: ignore[arg-type]

	def __ror__(self, other: bool) -> "Or[bool, S]":
		return Or(Const(None, other), self)  # type: ignore[arg-type]

	def __invert__(self) -> "Not[S]":
		return Not(self)  # type: ignore[arg-type]


class BinaryOperationOverloads(Expr[T, S, T]):
	@overload  # type: ignore[override]
	def __eq__(  # type: ignore[misc]  # pyright: ignore[reportOverlappingOverload]
		self: "ResultBase[TSupportsArithmetic, S]",
		other: "ConstTolerance[TSupportsArithmetic]",
	) -> "Result[TSupportsArithmetic, S, bool]": ...

	@overload
	def __eq__(
		self, other: "ConstTolerance[TSupportsArithmetic]"
	) -> "Approximately[TSupportsArithmetic, S]": ...

	@overload
	def __eq__(  # type: ignore[misc]  # pyright: ignore[reportOverlappingOverload]
		self: "ResultBase[TSupportsEquality, S]",
		other: TSupportsEquality | Expr[TSupportsEquality, S, Any],
	) -> "Result[TSupportsEquality, S, bool]": ...

	@overload
	def __eq__(  # type: ignore[overload-overlap]  # pyright: ignore[reportOverlappingOverload]
		self, other: "ResultBase[TSupportsEquality, S]"
	) -> "Result[TSupportsEquality, S, bool]": ...

	@overload
	def __eq__(
		self, other: Expr[TSupportsEquality, S, TSupportsEquality]
	) -> "Eq[TSupportsEquality, S]": ...

	@overload
	def __eq__(self, other: TSupportsEquality) -> "Eq[TSupportsEquality, S]": ...

	def __eq__(  # type: ignore[misc]  # pyright: ignore[reportIncompatibleMethodOverride]
		self,
		other: Expr[TSupportsEquality, S, TSupportsEquality]
		| TSupportsEquality
		| "ConstTolerance[TSupportsArithmetic]",
	) -> "Eq[TSupportsEquality, S] | Approximately[TSupportsArithmetic, S] | Result[TSupportsEquality, S, bool] | Result[TSupportsArithmetic, S, bool]":
		if isinstance(other, ConstTolerance):
			if isinstance(self, ResultBase):
				return Result(
					Approximately,
					self,  # pyright: ignore[reportUnknownArgumentType]
					Pure(other),  # pyright: ignore[reportUnknownArgumentType]
				)
			return Approximately(self, other)  # type: ignore[arg-type]
		rhs = other if isinstance(other, Expr) else Const(None, other)  # pyright: ignore[reportUnknownVariableType]
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Eq,
				self if isinstance(self, ResultBase) else Pure(self),  # pyright: ignore[reportUnknownArgumentType]
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
			)
		return Eq(self, rhs)  # type: ignore[arg-type]

	@overload  # type: ignore[override]
	def __ne__(  # type: ignore[misc]
		self: "ResultBase[T, S]",
		other: T | Expr[T, S, Any],
	) -> "Result[T, S, bool]": ...

	@overload
	def __ne__(self, other: "ResultBase[T, S]") -> "Result[T, S, bool]": ...  # type: ignore[overload-overlap]

	@overload
	def __ne__(self, other: Expr[T, S, T]) -> "Ne[T, S]": ...

	@overload
	def __ne__(self, other: T) -> "Ne[T, S]": ...

	def __ne__(  # pyright: ignore[reportIncompatibleMethodOverride]
		self, other: Expr[T, S, T] | T
	) -> "Ne[T, S] | Result[T, S, bool]":
		rhs = other if isinstance(other, Expr) else Const[T](None, other)  # pyright: ignore[reportUnknownVariableType]
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Ne,
				self if isinstance(self, ResultBase) else Pure(self),  # pyright: ignore[reportUnknownArgumentType]
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
			)
		return Ne(self, rhs)  # pyright: ignore[reportUnknownArgumentType]

	@overload
	def __lt__(  # type: ignore[misc]
		self: "ResultBase[TSupportsComparison, S]",
		other: TSupportsComparison | Expr[TSupportsComparison, S, Any],
	) -> "Result[TSupportsComparison, S, bool]": ...

	@overload
	def __lt__(self, other: TSupportsComparison) -> "Lt[TSupportsComparison, S]": ...

	@overload
	def __lt__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsComparison, S]"
	) -> "Result[TSupportsComparison, S, bool]": ...

	@overload
	def __lt__(
		self, other: Expr[TSupportsComparison, S, TSupportsComparison]
	) -> "Lt[TSupportsComparison, S]": ...

	def __lt__(
		self, other: Expr[TSupportsComparison, S, TSupportsComparison] | TSupportsComparison
	) -> "Lt[TSupportsComparison, S] | Result[TSupportsComparison, S, bool]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Lt,
				self if isinstance(self, ResultBase) else Pure(self),
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
			)
		return Lt(self, rhs)  # type: ignore[arg-type]

	@overload
	def __le__(  # type: ignore[misc]
		self: "ResultBase[TSupportsComparison, S]",
		other: TSupportsComparison | Expr[TSupportsComparison, S, Any],
	) -> "Result[TSupportsComparison, S, bool]": ...

	@overload
	def __le__(self, other: TSupportsComparison) -> "Le[TSupportsComparison, S]": ...

	@overload
	def __le__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsComparison, S]"
	) -> "Result[TSupportsComparison, S, bool]": ...

	@overload
	def __le__(
		self, other: Expr[TSupportsComparison, S, TSupportsComparison]
	) -> "Le[TSupportsComparison, S]": ...

	def __le__(
		self, other: Expr[TSupportsComparison, S, TSupportsComparison] | TSupportsComparison
	) -> "Le[TSupportsComparison, S] | Result[TSupportsComparison, S, bool]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Le,
				self if isinstance(self, ResultBase) else Pure(self),
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
			)
		return Le(self, rhs)  # type: ignore[arg-type]

	@overload
	def __gt__(  # type: ignore[misc]
		self: "ResultBase[TSupportsComparison, S]",
		other: TSupportsComparison | Expr[TSupportsComparison, S, Any],
	) -> "Result[TSupportsComparison, S, bool]": ...

	@overload
	def __gt__(self, other: TSupportsComparison) -> "Gt[TSupportsComparison, S]": ...

	@overload
	def __gt__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsComparison, S]"
	) -> "Result[TSupportsComparison, S, bool]": ...

	@overload
	def __gt__(
		self, other: Expr[TSupportsComparison, S, TSupportsComparison]
	) -> "Gt[TSupportsComparison, S]": ...

	def __gt__(
		self, other: Expr[TSupportsComparison, S, TSupportsComparison] | TSupportsComparison
	) -> "Gt[TSupportsComparison, S] | Result[TSupportsComparison, S, bool]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Gt,
				self if isinstance(self, ResultBase) else Pure(self),
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
			)
		return Gt(self, rhs)  # type: ignore[arg-type]

	@overload
	def __ge__(  # type: ignore[misc]
		self: "ResultBase[TSupportsComparison, S]",
		other: TSupportsComparison | Expr[TSupportsComparison, S, Any],
	) -> "Result[TSupportsComparison, S, bool]": ...

	@overload
	def __ge__(self, other: TSupportsComparison) -> "Ge[TSupportsComparison, S]": ...

	@overload
	def __ge__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsComparison, S]"
	) -> "Result[TSupportsComparison, S, bool]": ...

	@overload
	def __ge__(
		self, other: Expr[TSupportsComparison, S, TSupportsComparison]
	) -> "Ge[TSupportsComparison, S]": ...

	def __ge__(
		self, other: Expr[TSupportsComparison, S, TSupportsComparison] | TSupportsComparison
	) -> "Ge[TSupportsComparison, S] | Result[TSupportsComparison, S, bool]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Ge,
				self if isinstance(self, ResultBase) else Pure(self),
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
			)
		return Ge(self, rhs)  # type: ignore[arg-type]

	@overload
	def __add__(  # type: ignore[misc]
		self: "ResultBase[TSupportsArithmetic, S]",
		other: TSupportsArithmetic | Expr[TSupportsArithmetic, S, Any],
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __add__(self, other: TSupportsArithmetic) -> "Add[TSupportsArithmetic, S]": ...

	@overload
	def __add__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsArithmetic, S]"
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __add__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic]
	) -> "Add[TSupportsArithmetic, S]": ...

	def __add__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Add[TSupportsArithmetic, S] | Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Add,
				self if isinstance(self, ResultBase) else Pure(self),
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
			)
		return Add(self, rhs)  # type: ignore[arg-type]

	@overload
	def __radd__(  # type: ignore[misc]
		self: "ResultBase[TSupportsArithmetic, S]",
		other: TSupportsArithmetic | Expr[TSupportsArithmetic, S, Any],
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __radd__(self, other: TSupportsArithmetic) -> "Add[TSupportsArithmetic, S]": ...

	@overload
	def __radd__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsArithmetic, S]"
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __radd__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic]
	) -> "Add[TSupportsArithmetic, S]": ...

	def __radd__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Add[TSupportsArithmetic, S] | Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Add,
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
				self if isinstance(self, ResultBase) else Pure(self),
			)
		return Add(rhs, self)  # type: ignore[arg-type]

	@overload
	def __sub__(  # type: ignore[misc]
		self: "ResultBase[TSupportsArithmetic, S]",
		other: TSupportsArithmetic | Expr[TSupportsArithmetic, S, Any],
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __sub__(self, other: TSupportsArithmetic) -> "Sub[TSupportsArithmetic, S]": ...

	@overload
	def __sub__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsArithmetic, S]"
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __sub__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic]
	) -> "Sub[TSupportsArithmetic, S]": ...

	def __sub__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Sub[TSupportsArithmetic, S] | Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Sub,
				self if isinstance(self, ResultBase) else Pure(self),
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
			)
		return Sub(self, rhs)  # type: ignore[arg-type]

	@overload
	def __rsub__(  # type: ignore[misc]
		self: "ResultBase[TSupportsArithmetic, S]",
		other: TSupportsArithmetic | Expr[TSupportsArithmetic, S, Any],
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __rsub__(self, other: TSupportsArithmetic) -> "Sub[TSupportsArithmetic, S]": ...

	@overload
	def __rsub__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsArithmetic, S]"
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __rsub__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic]
	) -> "Sub[TSupportsArithmetic, S]": ...

	def __rsub__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Sub[TSupportsArithmetic, S] | Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Sub,
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
				self if isinstance(self, ResultBase) else Pure(self),
			)
		return Sub(rhs, self)  # type: ignore[arg-type]

	@overload
	def __mul__(  # type: ignore[misc]
		self: "ResultBase[TSupportsArithmetic, S]",
		other: TSupportsArithmetic | Expr[TSupportsArithmetic, S, Any],
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __mul__(self, other: TSupportsArithmetic) -> "Mul[TSupportsArithmetic, S]": ...

	@overload
	def __mul__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsArithmetic, S]"
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __mul__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic]
	) -> "Mul[TSupportsArithmetic, S]": ...

	def __mul__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Mul[TSupportsArithmetic, S] | Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Mul,
				self if isinstance(self, ResultBase) else Pure(self),
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
			)
		return Mul(self, rhs)  # type: ignore[arg-type]

	@overload
	def __rmul__(  # type: ignore[misc]
		self: "ResultBase[TSupportsArithmetic, S]",
		other: TSupportsArithmetic | Expr[TSupportsArithmetic, S, Any],
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __rmul__(self, other: TSupportsArithmetic) -> "Mul[TSupportsArithmetic, S]": ...

	@overload
	def __rmul__(  # type: ignore[overload-overlap]
		self, other: "ResultBase[TSupportsArithmetic, S]"
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]": ...

	@overload
	def __rmul__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic]
	) -> "Mul[TSupportsArithmetic, S]": ...

	def __rmul__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Mul[TSupportsArithmetic, S] | Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		if isinstance(self, ResultBase) or isinstance(rhs, ResultBase):
			return Result(
				Mul,
				rhs if isinstance(rhs, ResultBase) else Pure(rhs),
				self if isinstance(self, ResultBase) else Pure(self),
			)
		return Mul(rhs, self)  # type: ignore[arg-type]

	def __truediv__(self, other: Expr[float, S, Any] | float) -> "Result[float, S, float]":
		rhs = other if isinstance(other, Expr) else Const[float](None, other)
		return Result(
			Div,
			self if isinstance(self, ResultBase) else Pure(self),  # pyright: ignore[reportUnknownArgumentType]
			rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
		)

	def __rtruediv__(self, other: Expr[float, S, Any] | float) -> "Result[float, S, float]":
		rhs = other if isinstance(other, Expr) else Const[float](None, other)
		return Result(
			Div,
			rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
			self if isinstance(self, ResultBase) else Pure(self),  # pyright: ignore[reportUnknownArgumentType]
		)

	def __pow__(
		self, power: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = power if isinstance(power, Expr) else Const(None, power)
		return Result(
			Pow,
			self if isinstance(self, ResultBase) else Pure(self),  # pyright: ignore[reportUnknownArgumentType]
			rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
		)

	def __rpow__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		return Result(
			Pow,
			rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
			self if isinstance(self, ResultBase) else Pure(self),  # pyright: ignore[reportUnknownArgumentType]
		)

	def __mod__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		return Result(
			Mod,
			self if isinstance(self, ResultBase) else Pure(self),  # pyright: ignore[reportUnknownArgumentType]
			rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
		)

	def __rmod__(
		self, other: Expr[TSupportsArithmetic, S, TSupportsArithmetic] | TSupportsArithmetic
	) -> "Result[TSupportsArithmetic, S, TSupportsArithmetic]":
		rhs = other if isinstance(other, Expr) else Const(None, other)
		return Result(
			Mod,
			rhs if isinstance(rhs, ResultBase) else Pure(rhs),  # pyright: ignore[reportUnknownArgumentType]
			self if isinstance(self, ResultBase) else Pure(self),  # pyright: ignore[reportUnknownArgumentType]
		)

	def __neg__(self) -> "Neg[T, S]":
		return Neg(self)


@dataclass(frozen=True, eq=False, slots=True)
class Const(BinaryOperationOverloads[T, Any], BooleanBinaryOperationOverloads[T, Any]):
	"""A constant that evaluates to itself.

	>>> MY_CONST = Const("My Const", 42)
	>>> MY_CONST.eval(None)
	Const(name='My Const', value=42)
	>>> MY_CONST.to_string()
	'My Const:42'
	>>> MY_CONST.unwrap(None)
	42

	You can create an unnamed constant by passing `None` as the name, but in
	this case, literals should be preferred.

	>>> MY_UNNAMED_CONST = Const(None, 42)
	>>> MY_UNNAMED_CONST.to_string()
	'42'
	>>> MY_LITERAL_CONST = 42
	>>> str(MY_LITERAL_CONST)
	'42'
	>>> (MY_CONST == MY_UNNAMED_CONST).to_string({})
	'(My Const:42 == 42 -> True)'
	>>> (MY_CONST == 42).to_string({})
	'(My Const:42 == 42 -> True)'

	"""

	name: str | None
	value: T

	def eval(self, ctx: Any) -> "Const[T]":  # noqa: ARG002
		return self

	def to_string(self, ctx: Any | None = None) -> str:  # noqa: ARG002
		return f"{self.name}:{self.value}" if self.name else str(self.value)

	def partial(self, ctx: Any) -> "Const[T]":  # noqa: ARG002
		return self


@dataclass(frozen=True, eq=False, slots=True)
class UnaryOpEval(Eval[bool, S]):
	left: Expr[Any, S, Any]


class UnaryOpToString(ToString[S], UnaryOpEval[S]):
	op: ClassVar[str] = "not "
	template: ClassVar[str] = "({op}{left})"
	template_eval: ClassVar[str] = "({op}{left} -> {out})"

	def to_string(self, ctx: S | None = None) -> str:
		left: Final = self.left.to_string(ctx)
		if ctx is None:
			return self.template.format(op=self.op, left=left)
		else:
			return self.template_eval.format(op=self.op, left=left, out=self.eval(ctx).value)


class UnaryToStringMixin(ToString[S], Eval[T, S], Generic[T, S]):
	left: Expr[T, S, T]
	op: ClassVar[str]
	template: ClassVar[str]
	template_eval: ClassVar[str]

	def to_string(self, ctx: S | None = None) -> str:
		left: Final = self.left.to_string(ctx)
		if ctx is None:
			return self.template.format(op=self.op, left=left)
		return self.template_eval.format(op=self.op, left=left, out=self.eval(ctx).value)


class Not(UnaryOpToString[S], UnaryOperationOverloads[S], BooleanBinaryOperationOverloads[bool, S]):
	op: ClassVar[str] = "not "

	def eval(self, ctx: S) -> Const[bool]:
		return Const(None, not self.left.eval(ctx).value)

	def partial(self, ctx: Any) -> "Expr[bool, Any, bool]":
		return Not(self.left.partial(ctx))


@dataclass(frozen=True, eq=False, slots=True)
class Neg(
	UnaryToStringMixin[T, S],
	BinaryOperationOverloads[T, S],
	BooleanBinaryOperationOverloads[T, S],
):
	left: Expr[T, S, T]
	op: ClassVar[str] = "-"
	template: ClassVar[str] = "({op}{left})"
	template_eval: ClassVar[str] = "({op}{left} -> {out})"

	def eval(self, ctx: S) -> Const[T]:
		return Const(None, -self.left.eval(ctx).value)  # type: ignore[operator]

	def partial(self, ctx: Any) -> "Neg[T, Any]":
		return Neg(self.left.partial(ctx))


class BinaryOpProtocol(Expr[T, S, R], Protocol[T, S, R]):
	op_func: ClassVar[Callable[..., Any]]
	left: Expr[T, S, T]
	right: Expr[T, S, T]


@dataclass(frozen=True, eq=False, slots=True)
class BinaryOp(ToString[S], BinaryOpProtocol[T, S, R], Generic[T, S, R]):
	op: ClassVar[str] = " ? "
	template: ClassVar[str] = "({left}{op}{right})"
	template_eval: ClassVar[str] = "({left}{op}{right} -> {out})"

	left: Expr[T, S, T]
	right: Expr[T, S, T]

	def eval(self, ctx: S) -> Const[R]:
		return Const(None, self.op_func(self.left.unwrap(ctx), self.right.unwrap(ctx)))

	def to_string(self, ctx: S | None = None) -> str:
		left: Final = self.left.to_string(ctx)
		right: Final = self.right.to_string(ctx)
		if ctx is None:
			return self.template.format(left=left, op=self.op, right=right)
		else:
			return self.template_eval.format(
				left=left, op=self.op, right=right, out=self.eval(ctx).value
			)

	def partial(self, ctx: Any) -> "Expr[T, Any, R]":
		return type(self)(self.left.partial(ctx), self.right.partial(ctx))


class SymmetricalBinaryOpProtocol(BinaryOpProtocol[T, S, T], Protocol[T, S]):
	identity_element: ClassVar[Any]


@dataclass(frozen=True, eq=False, slots=True)
class Foldable(SymmetricalBinaryOpProtocol[T, S], BinaryOp[T, S, T]): ...


class And(
	Foldable[TSupportsLogic, S],
	BooleanBinaryOperationOverloads[TSupportsLogic, S],
):
	op: ClassVar[str] = " & "
	identity_element: ClassVar = True

	@staticmethod
	def op_func(a: TSupportsLogic, b: TSupportsLogic) -> TSupportsLogic:
		return a and b


class Or(
	Foldable[TSupportsLogic, S],
	BooleanBinaryOperationOverloads[TSupportsLogic, S],
):
	op: ClassVar[str] = " | "
	identity_element: ClassVar = False

	@staticmethod
	def op_func(a: TSupportsLogic, b: TSupportsLogic) -> TSupportsLogic:
		return a or b


class Eq(  # pyright: ignore[reportGeneralTypeIssues]
	BinaryOp[TSupportsEquality, S, bool],
	BinaryOperationOverloads[TSupportsEquality, S],
	BooleanBinaryOperationOverloads[bool, S],
):
	op: ClassVar[str] = " == "
	op_func: ClassVar = operator.eq  # pyright: ignore[reportInvalidTypeForm]


class Ne(  # pyright: ignore[reportGeneralTypeIssues]
	BinaryOp[TSupportsEquality, S, bool],
	BinaryOperationOverloads[TSupportsEquality, S],
	BooleanBinaryOperationOverloads[bool, S],
):
	op: ClassVar[str] = " != "
	op_func: ClassVar = operator.ne  # pyright: ignore[reportInvalidTypeForm]


class Lt(  # pyright: ignore[reportGeneralTypeIssues]
	BinaryOp[TSupportsComparison, S, bool],
	BinaryOperationOverloads[TSupportsComparison, S],
	BooleanBinaryOperationOverloads[bool, S],
):
	op: ClassVar[str] = " < "
	op_func: ClassVar = operator.lt  # pyright: ignore[reportInvalidTypeForm]


class Le(  # pyright: ignore[reportGeneralTypeIssues]
	BinaryOp[TSupportsComparison, S, bool],
	BinaryOperationOverloads[TSupportsComparison, S],
	BooleanBinaryOperationOverloads[bool, S],
):
	op: ClassVar[str] = " <= "
	op_func: ClassVar = operator.le  # pyright: ignore[reportInvalidTypeForm]


class Gt(  # pyright: ignore[reportGeneralTypeIssues]
	BinaryOp[TSupportsComparison, S, bool],
	BinaryOperationOverloads[TSupportsComparison, S],
	BooleanBinaryOperationOverloads[bool, S],
):
	op: ClassVar[str] = " > "
	op_func: ClassVar = operator.gt  # pyright: ignore[reportInvalidTypeForm]


class Ge(  # pyright: ignore[reportGeneralTypeIssues]
	BinaryOp[TSupportsComparison, S, bool],
	BinaryOperationOverloads[TSupportsComparison, S],
	BooleanBinaryOperationOverloads[bool, S],
):
	op: ClassVar[str] = " >= "
	op_func: ClassVar = operator.ge  # pyright: ignore[reportInvalidTypeForm]


class Min(
	Foldable[TSupportsComparison, S],
	BinaryOperationOverloads[TSupportsComparison, S],
):
	op: ClassVar[str] = "min"
	template: ClassVar[str] = "({op} {left} {right})"
	template_eval: ClassVar[str] = "({op} {left} {right} -> {out})"
	op_func: ClassVar = min
	identity_element: ClassVar = float("inf")


class Max(
	Foldable[TSupportsComparison, S],
	BinaryOperationOverloads[TSupportsComparison, S],
):
	op: ClassVar[str] = "max"
	template: ClassVar[str] = "({op} {left} {right})"
	template_eval: ClassVar[str] = "({op} {left} {right} -> {out})"
	op_func: ClassVar = max
	identity_element: ClassVar = float("-inf")


class Add(
	Foldable[TSupportsArithmetic, S],
	BinaryOperationOverloads[TSupportsArithmetic, S],
):
	op: ClassVar[str] = " + "
	op_func: ClassVar = operator.add
	identity_element: ClassVar = 0


class Sub(
	Foldable[TSupportsArithmetic, S],
	BinaryOperationOverloads[TSupportsArithmetic, S],
):
	op: ClassVar[str] = " - "
	op_func: ClassVar = operator.sub
	identity_element: ClassVar = 0


class Mul(
	Foldable[TSupportsArithmetic, S],
	BinaryOperationOverloads[TSupportsArithmetic, S],
):
	op: ClassVar[str] = " * "
	op_func: ClassVar = operator.mul
	identity_element: ClassVar = 1


class Div(
	Foldable[TSupportsArithmetic, S],
	BinaryOperationOverloads[TSupportsArithmetic, S],
):
	op: ClassVar[str] = " / "
	op_func: ClassVar = operator.truediv
	identity_element: ClassVar = 1


@dataclass(frozen=True, eq=False, slots=True)
class Pow(
	Foldable[TSupportsArithmetic, S],
	BinaryOperationOverloads[TSupportsArithmetic, S],
):
	op: ClassVar[str] = "^"
	op_func: ClassVar = operator.pow
	identity_element: ClassVar = 1
	left: Expr[TSupportsArithmetic, S, TSupportsArithmetic]
	right: Expr[TSupportsArithmetic, S, TSupportsArithmetic]


@dataclass(frozen=True, eq=False, slots=True)
class Mod(
	BinaryOp[TSupportsArithmetic, S, TSupportsArithmetic],
	BinaryOperationOverloads[TSupportsArithmetic, S],
):
	op: ClassVar[str] = " % "
	op_func: ClassVar = operator.mod

	left: Expr[TSupportsArithmetic, S, TSupportsArithmetic]
	right: Expr[TSupportsArithmetic, S, TSupportsArithmetic]


class ResultBase[T, S](Expr[T, S, T]):
	"""Type class marker for the Validated applicative (error-accumulating) context.

	Statically, ResultBase[T, S] is Expr[T, S, T] — Failure is a runtime concern.
	BinaryOperationOverloads uses ``isinstance(x, ResultBase)`` to decide
	whether an operator returns a clean op (Add, Eq, ...) or a Result that
	accumulates Failures from both operands.

	>>> from typing import NamedTuple
	>>> from mahonia import ResultBase, Var
	>>> from mahonia.python_func import python_func
	>>> class Ctx(NamedTuple):
	... 	x: float

	Plain expressions are not in the Result context:

	>>> x = Var[float, Ctx]("x")
	>>> isinstance(x, ResultBase)
	False

	python_func lifts a Python function into the Result context:

	>>> safe_abs = python_func(abs)
	>>> isinstance(safe_abs(x), ResultBase)
	True
	"""

	def unwrap(self, ctx: S) -> T | Failure:  # type: ignore[override]
		result = self.eval(ctx)
		return result if isinstance(result, Failure) else result.value


@dataclass(frozen=True, eq=False, slots=True)
class Pure[T, S](  # pyright: ignore[reportGeneralTypeIssues]
	ResultBase[T, S],
	BinaryOperationOverloads[T, S],
	BooleanBinaryOperationOverloads[T, S],
):
	"""Applicative pure: lifts a clean Expr[T, S, T] into ResultBase[T, S].

	Pure can never produce a Failure — it is the identity embedding.
	Transparent in serialization: Pure(x).to_string() == x.to_string().

	>>> from typing import NamedTuple
	>>> from mahonia import Var, Pure, Result
	>>> class Ctx(NamedTuple):
	... 	x: float
	... 	y: float
	>>> x, y = Var[float, Ctx]("x"), Var[float, Ctx]("y")

	Wrapping in Pure is invisible to evaluation and serialization:

	>>> Pure(x).to_string()
	'x'
	>>> Pure(x).unwrap(Ctx(x=42.0, y=0.0))
	42.0

	But it marks the expression as ResultBase, so operators produce Result:

	>>> type(x + y).__name__
	'Add'
	>>> type(Pure(x) + y).__name__
	'Result'
	"""

	inner: Expr[T, S, T]

	def eval(self, ctx: S) -> Const[T]:
		return self.inner.eval(ctx)

	def to_string(self, ctx: S | None = None) -> str:
		return self.inner.to_string(ctx)

	def partial(self, ctx: Any) -> "Pure[T, Any]":  # pyright: ignore[reportIncompatibleMethodOverride]
		return Pure(self.inner.partial(ctx))


@dataclass(frozen=True, eq=False, slots=True)
class Result[T, S, R](  # pyright: ignore[reportGeneralTypeIssues]
	ResultBase[R, S],
	BinaryOperationOverloads[R, S],
	BooleanBinaryOperationOverloads[R, S],
	ToString[S],
):
	"""liftA2 for the Validated applicative: applies a BinaryOp in the error-accumulating context.

	When both operands succeed, applies inner.op_func.  When both fail, the
	Failures are concatenated.  Typically constructed via operators on
	ResultBase values rather than directly.

	>>> from typing import NamedTuple
	>>> from mahonia import Failure, Var
	>>> from mahonia.python_func import python_func
	>>> class Ctx(NamedTuple):
	... 	x: float
	... 	y: float
	>>> x, y = Var[float, Ctx]("x"), Var[float, Ctx]("y")
	>>> def my_sqrt(v: float) -> float:
	... 	if v < 0: raise ValueError(f"neg: {v}")
	... 	return v ** 0.5
	>>> safe_sqrt = python_func(my_sqrt)
	>>> expr = safe_sqrt(x) + safe_sqrt(y)

	Both succeed — the underlying Add is applied:

	>>> expr.to_string(Ctx(x=4.0, y=9.0))
	'(my_sqrt(x:4.0) -> 2.0 + my_sqrt(y:9.0) -> 3.0 -> 5.0)'

	Both fail — exceptions are accumulated, not short-circuited:

	>>> expr.to_string(Ctx(x=-1.0, y=-4.0))
	"(my_sqrt(x:-1.0) -> Failure(exceptions=(ValueError('neg: -1.0'),)) + my_sqrt(y:-4.0) -> Failure(exceptions=(ValueError('neg: -4.0'),)) -> Failure(exceptions=(ValueError('neg: -1.0'), ValueError('neg: -4.0'))))"
	"""

	inner: type[BinaryOp[Any, Any, Any]]
	left: ResultBase[Any, S]
	right: ResultBase[Any, S]

	def eval(self, ctx: S) -> Const[R]:
		lv: Final = self.left.unwrap(ctx)
		rv: Final = self.right.unwrap(ctx)
		match (lv, rv):
			case (Failure() as f1, Failure() as f2):
				return Failure(f1.exceptions + f2.exceptions)  # type: ignore[return-value]
			case (Failure() as f, _) | (_, Failure() as f):
				return f  # type: ignore[return-value]
			case _:
				try:
					return Const(None, self.inner.op_func(lv, rv))
				except Exception as e:
					return Failure((e,))  # type: ignore[return-value]

	def to_string(self, ctx: S | None = None) -> str:
		left: Final = self.left.to_string(ctx)
		right: Final = self.right.to_string(ctx)
		if ctx is None:
			return self.inner.template.format(left=left, op=self.inner.op, right=right)
		return self.inner.template_eval.format(
			left=left, op=self.inner.op, right=right, out=self.unwrap(ctx)
		)

	def partial(self, ctx: Any) -> "Result[T, Any, R]":  # pyright: ignore[reportIncompatibleMethodOverride]
		return Result(self.inner, self.left.partial(ctx), self.right.partial(ctx))  # type: ignore[arg-type]


from mahonia.bound import BoundExpr as BoundExpr  # noqa: E402
from mahonia.containers import AllExpr as AllExpr  # noqa: E402
from mahonia.containers import AnyExpr as AnyExpr  # noqa: E402
from mahonia.containers import BoolContainerFoldBase as BoolContainerFoldBase  # noqa: E402
from mahonia.containers import (  # noqa: E402
	ComparableContainerFoldBase as ComparableContainerFoldBase,
)
from mahonia.containers import Contains as Contains  # noqa: E402
from mahonia.containers import FilterExpr as FilterExpr  # noqa: E402
from mahonia.containers import FoldLExpr as FoldLExpr  # noqa: E402
from mahonia.containers import MapExpr as MapExpr  # noqa: E402
from mahonia.containers import MaxExpr as MaxExpr  # noqa: E402
from mahonia.containers import MinExpr as MinExpr  # noqa: E402
from mahonia.context import context_vars as context_vars  # noqa: E402
from mahonia.context import merge as merge  # noqa: E402
from mahonia.extract_vars import extract_vars as extract_vars  # noqa: E402
from mahonia.formatting import format_iterable_var as format_iterable_var  # noqa: E402
from mahonia.match import Match as Match  # noqa: E402
from mahonia.match import MatchExpr as MatchExpr  # noqa: E402
from mahonia.predicate import Predicate as Predicate  # noqa: E402
from mahonia.tolerance import Approximately as Approximately  # noqa: E402
from mahonia.tolerance import ConstTolerance as ConstTolerance  # noqa: E402
from mahonia.tolerance import ConstToleranceProtocol as ConstToleranceProtocol  # noqa: E402
from mahonia.tolerance import Percent as Percent  # noqa: E402
from mahonia.tolerance import PlusMinus as PlusMinus  # noqa: E402
from mahonia.unary import Abs as Abs  # noqa: E402
from mahonia.unary import Clamp as Clamp  # noqa: E402
from mahonia.unary import ClampExpr as ClampExpr  # noqa: E402
from mahonia.var import Var as Var  # noqa: E402

type FloatVar[S] = Var[float, S]
type IntVar[S] = Var[int, S]
type BoolVar[S] = Var[bool, S]
type StrVar[S] = Var[str, S]
type ListVar[T, S] = Var[SizedIterable[T], S]



# Copyright (c) 2025 JP Hutchins
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from typing import Any, ClassVar, Final, Generic, overload

from mahonia import (
	And,
	BinaryOperationOverloads,
	BooleanBinaryOperationOverloads,
	Const,
	Expr,
	Foldable,
	Func,
	Max,
	Min,
	Or,
	R,
	S,
	SizedIterable,
	T,
	TSupportsComparison,
	U,
)
from mahonia.formatting import format_iterable_var


@dataclass(frozen=True, eq=False, slots=True)
class Contains(
	BinaryOperationOverloads[bool, S],
	BooleanBinaryOperationOverloads[bool, S],
	Generic[T, S],
):
	"""Check if a value is contained in a collection.

	>>> from typing import NamedTuple
	>>> from mahonia import Var, SizedIterable
	>>> class Ctx(NamedTuple):
	... 	values: list[int]
	... 	target: int
	>>> values = Var[SizedIterable[int], Ctx]("values")
	>>> target = Var[int, Ctx]("target")
	>>> contains_expr = Contains(target, values)
	>>> contains_expr.to_string()
	'(target in values)'
	>>> contains_expr.unwrap(Ctx(values=[1, 2, 3], target=2))
	True
	>>> contains_expr.unwrap(Ctx(values=[1, 2, 3], target=5))
	False
	>>> contains_expr.to_string(Ctx(values=[1, 2, 3], target=2))
	'(target:2 in values:3[1,..3] -> True)'
	"""

	op: ClassVar[str] = " in "
	template: ClassVar[str] = "({left}{op}{right})"
	template_eval: ClassVar[str] = "({left}{op}{right} -> {out})"

	element: Expr[T, S, T]
	container: Expr[SizedIterable[T], S, SizedIterable[T]]

	def eval(self, ctx: S) -> Const[bool]:
		return Const(None, self.element.unwrap(ctx) in self.container.unwrap(ctx))

	def unwrap(self, ctx: S) -> bool:
		return self.eval(ctx).value

	def to_string(self, ctx: S | None = None) -> str:
		left: Final = self.element.to_string(ctx)
		right: Final = format_iterable_var(self.container, ctx)
		if ctx is None:
			return self.template.format(left=left, op=self.op, right=right)
		else:
			return self.template_eval.format(
				left=left, op=self.op, right=right, out=self.eval(ctx).value
			)

	def partial(self, ctx: Any) -> "Expr[bool, Any, bool]":
		return Contains(self.element.partial(ctx), self.container.partial(ctx))


@dataclass(frozen=True, eq=False, slots=True)
class MapExpr(
	BinaryOperationOverloads[SizedIterable[U], S],
	Generic[T, U, S],
):
	"""Apply a function to each element in a container."""

	op: ClassVar[str] = "map"
	template: ClassVar[str] = "({op} {func} {container})"
	template_eval: ClassVar[str] = "({op} {func} {container} -> {out})"

	func: "Func[U, Any]"
	container: Expr[SizedIterable[T], S, SizedIterable[T]]

	def eval(self, ctx: S) -> "Const[SizedIterable[U]]":
		container_values = self.container.unwrap(ctx)
		result: list[U] = []
		for item in container_values:
			if self.func.args and hasattr(self.func.args[0], "name"):
				arg_name = self.func.args[0].name  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType, reportUnknownVariableType]
				temp_ctx = type("TempCtx", (), {arg_name: item})()
				result.append(self.func.expr.unwrap(temp_ctx))  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
		return Const(None, result)

	def unwrap(self, ctx: S) -> SizedIterable[U]:
		return self.eval(ctx).value

	def to_string(self, ctx: S | None = None) -> str:
		func_str = self.func.to_string()
		if ctx is None:
			container_str = self.container.to_string()
			return self.template.format(op=self.op, func=func_str, container=container_str)
		else:
			container_str = format_iterable_var(self.container, ctx)
			result_value = self.eval(ctx).value
			result_expr = Const(None, result_value)
			out_str = format_iterable_var(result_expr, ctx)
			return self.template_eval.format(
				op=self.op, func=func_str, container=container_str, out=out_str
			)

	def partial(self, ctx: Any) -> "Expr[SizedIterable[U], Any, SizedIterable[U]]":
		partial_func = Func(self.func.args, self.func.expr.partial(ctx))
		return MapExpr(partial_func, self.container.partial(ctx))


@dataclass(frozen=True, eq=False, slots=True)
class FoldLExpr(
	BinaryOperationOverloads[R, S],
	Generic[T, S, R],
):
	op_cls: type[Foldable[Any, Any]]
	container: Expr[SizedIterable[T], S, SizedIterable[T]]
	initial: R | None = None

	@overload
	def __init__(
		self: "FoldLExpr[Expr[Any, Any, R], S, R]",  # pyright: ignore[reportInvalidTypeVarUse]
		op_cls: type[Foldable[Any, Any]],
		container: Expr[SizedIterable[Expr[Any, Any, R]], S, SizedIterable[Expr[Any, Any, R]]],
		initial: None = None,
	) -> None: ...

	@overload
	def __init__(
		self: "FoldLExpr[Expr[Any, Any, R], S, R]",  # pyright: ignore[reportInvalidTypeVarUse]
		op_cls: type[Foldable[Any, Any]],
		container: Expr[SizedIterable[Expr[Any, Any, R]], S, SizedIterable[Expr[Any, Any, R]]],
		initial: R = ...,
	) -> None: ...

	@overload
	def __init__(
		self: "FoldLExpr[T, S, T]",  # pyright: ignore[reportInvalidTypeVarUse]
		op_cls: type[Foldable[Any, Any]],
		container: Expr[SizedIterable[T], S, SizedIterable[T]],
		initial: None = None,
	) -> None: ...

	@overload
	def __init__(
		self: "FoldLExpr[T, S, T]",  # pyright: ignore[reportInvalidTypeVarUse]
		op_cls: type[Foldable[Any, Any]],
		container: Expr[SizedIterable[T], S, SizedIterable[T]],
		initial: T = ...,
	) -> None: ...

	def __init__(
		self,
		op_cls: type[Foldable[Any, Any]],
		container: Expr[SizedIterable[Any], S, SizedIterable[Any]],
		initial: Any | None = None,
	) -> None:
		object.__setattr__(self, "op_cls", op_cls)
		object.__setattr__(self, "container", container)
		object.__setattr__(self, "initial", initial)

	def eval(self, ctx: S) -> Const[R]:
		result_value: R = self.initial if self.initial is not None else self.op_cls.identity_element  # pyright: ignore[reportGeneralTypeIssues]

		for item in self.container.unwrap(ctx):
			item_value: R = item.unwrap(ctx) if isinstance(item, Expr) else item  # type: ignore[assignment]  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType, reportAssignmentType]
			result_value = self.op_cls.op_func(result_value, item_value)  # pyright: ignore[reportArgumentType]

		return Const(None, result_value)

	def unwrap(self, ctx: S) -> R:
		return self.eval(ctx).value

	def to_string(self, ctx: S | None = None) -> str:
		op_str: Final = self.op_cls.op.strip()
		container_name: Final = (
			self.container.to_string()
			if ctx is None
			else self.container.to_string() + f":{len(self.container.unwrap(ctx))}"
		)
		if ctx is None:
			return f"(foldl {op_str} {container_name})"
		result = self.eval(ctx).value
		if self.op_cls in (Min, Max):
			return f"(foldl {op_str} {container_name} -> {result})"
		items = list(self.container.unwrap(ctx))
		initial_str = f"{str(self.initial)}{self.op_cls.op}" if self.initial is not None else ""
		items_str = self._format_items(items, op_str, ctx)
		return f"(foldl {op_str} {container_name} -> ({initial_str}{items_str}) -> {result})"

	@staticmethod
	def _format_items(items: list[Any], op: str, ctx: S) -> str:
		def serialize(item: Any) -> str:
			if isinstance(item, Expr):
				return item.to_string(ctx)  # pyright: ignore[reportUnknownMemberType]
			return str(item)

		return f" {op} ".join(serialize(i) for i in items)

	def partial(self, ctx: Any) -> "Expr[R, Any, R]":
		return FoldLExpr(self.op_cls, self.container.partial(ctx), self.initial)  # type: ignore[arg-type]


@dataclass(frozen=True, eq=False, slots=True)
class BoolContainerFoldBase(
	BinaryOperationOverloads[bool, S],
	BooleanBinaryOperationOverloads[bool, S],
):
	op: ClassVar[str]
	op_cls: ClassVar[type[Foldable[Any, Any]]]
	template: ClassVar[str] = "({op} {left})"
	template_eval: ClassVar[str] = "({op} {left} -> {out})"

	container: Expr[SizedIterable[Any], S, SizedIterable[Any]]
	_foldl: FoldLExpr[bool, S, bool] = field(init=False)

	def __post_init__(self) -> None:
		object.__setattr__(self, "_foldl", FoldLExpr(self.op_cls, self.container))

	def eval(self, ctx: S) -> Const[bool]:
		return self._foldl.eval(ctx)

	def unwrap(self, ctx: S) -> bool:
		return self._foldl.unwrap(ctx)

	def to_string(self, ctx: S | None = None) -> str:
		from mahonia import Var

		if ctx is None:
			return self.template.format(op=self.op, left=format_iterable_var(self.container, ctx))
		left = (
			format_iterable_var(self.container, ctx)
			if isinstance(self.container, (Var, Const))
			else self.container.to_string(ctx)
		)
		return self.template_eval.format(op=self.op, left=left, out=self.eval(ctx).value)


@dataclass(frozen=True, eq=False, slots=True)
class AnyExpr(BoolContainerFoldBase[S]):
	"""True if any element in the container is truthy.

	>>> from typing import NamedTuple
	>>> from mahonia import Var
	>>> class Ctx(NamedTuple):
	... 	values: list[bool]
	>>> values = Var[list[bool], Ctx]("values")
	>>> any_expr = AnyExpr(values)
	>>> any_expr.to_string()
	'(any values)'
	>>> any_expr.unwrap(Ctx(values=[False, True, False]))
	True
	>>> any_expr.unwrap(Ctx(values=[False, False, False]))
	False
	>>> any_expr.to_string(Ctx(values=[False, True, False]))
	'(any values:3[False,..False] -> True)'

	With complex expressions like MapExpr, shows the full evaluation trace:

	>>> class NumCtx(NamedTuple):
	... 	nums: list[int]
	>>> nums = Var[list[int], NumCtx]("nums")
	>>> n = Var[int, NumCtx]("n")
	>>> gt_five = (n > 5).map(nums)
	>>> any_gt_five = AnyExpr(gt_five)
	>>> any_gt_five.to_string()
	'(any (map n -> (n > 5) nums))'
	>>> any_gt_five.to_string(NumCtx(nums=[3, 7, 2]))
	'(any (map n -> (n > 5) nums:3[3,..2] -> 3[False,..False]) -> True)'
	>>> any_gt_five.to_string(NumCtx(nums=[1, 2, 3]))
	'(any (map n -> (n > 5) nums:3[1,..3] -> 3[False,..False]) -> False)'
	"""

	op: ClassVar[str] = "any"
	op_cls: ClassVar[type[Foldable[Any, Any]]] = Or

	def partial(self, ctx: Any) -> "AnyExpr[Any]":
		return AnyExpr(self.container.partial(ctx))


@dataclass(frozen=True, eq=False, slots=True)
class AllExpr(BoolContainerFoldBase[S]):
	"""True if all elements in the container are truthy.

	>>> from typing import NamedTuple
	>>> from mahonia import Var
	>>> class Ctx(NamedTuple):
	... 	values: list[bool]
	>>> values = Var[list[bool], Ctx]("values")
	>>> all_expr = AllExpr(values)
	>>> all_expr.to_string()
	'(all values)'
	>>> all_expr.unwrap(Ctx(values=[True, True, True]))
	True
	>>> all_expr.unwrap(Ctx(values=[True, False, True]))
	False
	>>> all_expr.to_string(Ctx(values=[True, False, True]))
	'(all values:3[True,..True] -> False)'

	With complex expressions like MapExpr, shows the full evaluation trace:

	>>> class NumCtx(NamedTuple):
	... 	nums: list[int]
	>>> nums = Var[list[int], NumCtx]("nums")
	>>> n = Var[int, NumCtx]("n")
	>>> lt_ten = (n < 10).map(nums)
	>>> all_lt_ten = AllExpr(lt_ten)
	>>> all_lt_ten.to_string()
	'(all (map n -> (n < 10) nums))'
	>>> all_lt_ten.to_string(NumCtx(nums=[3, 7, 2]))
	'(all (map n -> (n < 10) nums:3[3,..2] -> 3[True,..True]) -> True)'
	>>> all_lt_ten.to_string(NumCtx(nums=[3, 15, 2]))
	'(all (map n -> (n < 10) nums:3[3,..2] -> 3[True,..True]) -> False)'
	"""

	op: ClassVar[str] = "all"
	op_cls: ClassVar[type[Foldable[Any, Any]]] = And

	def partial(self, ctx: Any) -> "AllExpr[Any]":
		return AllExpr(self.container.partial(ctx))


@dataclass(frozen=True, eq=False, slots=True)
class ComparableContainerFoldBase(
	BinaryOperationOverloads[TSupportsComparison, S],
	Generic[TSupportsComparison, S],
):
	op: ClassVar[str]
	op_cls: ClassVar[type[Foldable[Any, Any]]]
	template: ClassVar[str] = "({op} {left})"
	template_eval: ClassVar[str] = "({op} {left} -> {out})"

	container: Expr[SizedIterable[TSupportsComparison], S, SizedIterable[TSupportsComparison]]
	_foldl: FoldLExpr[TSupportsComparison, S, TSupportsComparison] = field(init=False)

	def __post_init__(self) -> None:
		object.__setattr__(self, "_foldl", FoldLExpr(self.op_cls, self.container))

	def eval(self, ctx: S) -> Const[TSupportsComparison]:
		return self._foldl.eval(ctx)

	def unwrap(self, ctx: S) -> TSupportsComparison:
		return self._foldl.unwrap(ctx)

	def to_string(self, ctx: S | None = None) -> str:
		from mahonia import Var

		if ctx is None:
			return self.template.format(op=self.op, left=format_iterable_var(self.container, ctx))
		left = (
			format_iterable_var(self.container, ctx)
			if isinstance(self.container, (Var, Const))
			else self.container.to_string(ctx)
		)
		return self.template_eval.format(op=self.op, left=left, out=self.eval(ctx).value)


@dataclass(frozen=True, eq=False, slots=True)
class MinExpr(ComparableContainerFoldBase[TSupportsComparison, S]):
	"""Minimum element in a container.

	>>> from typing import NamedTuple
	>>> from mahonia import Var
	>>> class Ctx(NamedTuple):
	... 	values: list[int]
	>>> values = Var[list[int], Ctx]("values")
	>>> min_expr = MinExpr(values)
	>>> min_expr.to_string()
	'(min values)'
	>>> min_expr.unwrap(Ctx(values=[3, 1, 4, 1, 5]))
	1
	>>> min_expr.to_string(Ctx(values=[3, 1, 4]))
	'(min values:3[3,..4] -> 1)'
	"""

	op: ClassVar[str] = "min"
	op_cls: ClassVar[type[Foldable[Any, Any]]] = Min

	def partial(self, ctx: Any) -> "MinExpr[TSupportsComparison, Any]":
		return MinExpr(self.container.partial(ctx))


@dataclass(frozen=True, eq=False, slots=True)
class MaxExpr(ComparableContainerFoldBase[TSupportsComparison, S]):
	"""Maximum element in a container.

	>>> from typing import NamedTuple
	>>> from mahonia import Var
	>>> class Ctx(NamedTuple):
	... 	values: list[int]
	>>> values = Var[list[int], Ctx]("values")
	>>> max_expr = MaxExpr(values)
	>>> max_expr.to_string()
	'(max values)'
	>>> max_expr.unwrap(Ctx(values=[3, 1, 4, 1, 5]))
	5
	>>> max_expr.to_string(Ctx(values=[3, 1, 4]))
	'(max values:3[3,..4] -> 4)'
	"""

	op: ClassVar[str] = "max"
	op_cls: ClassVar[type[Foldable[Any, Any]]] = Max

	def partial(self, ctx: Any) -> "MaxExpr[TSupportsComparison, Any]":
		return MaxExpr(self.container.partial(ctx))


@dataclass(frozen=True, eq=False, slots=True)
class FilterExpr(
	BinaryOperationOverloads[SizedIterable[T], S],
	Generic[T, S],
):
	"""Filter container elements by predicate.

	>>> from typing import NamedTuple
	>>> from mahonia import Var, SizedIterable
	>>> class Ctx(NamedTuple):
	... 	values: list[int]
	>>> values = Var[SizedIterable[int], Ctx]("values")
	>>> x = Var[int, Ctx]("x")
	>>> is_positive = (x > 0).to_func()
	>>> filter_expr = FilterExpr(is_positive, values)
	>>> filter_expr.to_string()
	'(filter x -> (x > 0) values)'
	>>> filter_expr.unwrap(Ctx(values=[-1, 2, -3, 4, 5]))
	(2, 4, 5)
	>>> filter_expr.to_string(Ctx(values=[-1, 2, -3, 4, 5]))
	'(filter x -> (x > 0) values:5[-1,..5] -> 3[2,..5])'
	"""

	op: ClassVar[str] = "filter"
	template: ClassVar[str] = "({op} {func} {container})"
	template_eval: ClassVar[str] = "({op} {func} {container} -> {out})"

	predicate: Func[bool, Any]
	container: Expr[SizedIterable[T], S, SizedIterable[T]]

	def eval(self, ctx: S) -> Const[SizedIterable[T]]:
		predicate_results = MapExpr(self.predicate, self.container).unwrap(ctx)
		return Const(
			None,
			tuple(
				value
				for value, keep in zip(self.container.unwrap(ctx), predicate_results, strict=True)
				if keep
			),
		)

	def unwrap(self, ctx: S) -> SizedIterable[T]:
		return self.eval(ctx).value

	def to_string(self, ctx: S | None = None) -> str:
		func_str = self.predicate.to_string()
		if ctx is None:
			container_str = self.container.to_string()
			return self.template.format(op=self.op, func=func_str, container=container_str)
		else:
			container_str = format_iterable_var(self.container, ctx)
			result_value = self.eval(ctx).value
			result_expr = Const(None, result_value)
			out_str = format_iterable_var(result_expr, ctx)
			return self.template_eval.format(
				op=self.op, func=func_str, container=container_str, out=out_str
			)

	def partial(self, ctx: Any) -> "Expr[SizedIterable[T], Any, SizedIterable[T]]":
		partial_func = Func(self.predicate.args, self.predicate.expr.partial(ctx))
		return FilterExpr(partial_func, self.container.partial(ctx))



import inspect
from dataclasses import dataclass
from typing import Any, Callable, overload

from mahonia import (
	BinaryOperationOverloads,
	BooleanBinaryOperationOverloads,
	Const,
	Expr,
	Failure,
	ResultBase,
)
from mahonia.types import ContextProtocol


def _func_name(func: Callable[..., Any]) -> str:
	return func.__name__ if func.__name__ != "<lambda>" else "lambda"


class PythonFuncBase[R, S: ContextProtocol](  # pyright: ignore[reportGeneralTypeIssues]
	ResultBase[R, S],
	BinaryOperationOverloads[R, S],
	BooleanBinaryOperationOverloads[R, S],
):
	"""FFI bridge: wraps a Python callable into the Result context.

	Exceptions raised by the wrapped function become Failure values.
	When multiple arguments are themselves ResultBase, their Failures
	are accumulated before the function is called.

	>>> from typing import NamedTuple
	>>> from mahonia import Failure, Var
	>>> from mahonia.python_func import python_func
	>>> class Ctx(NamedTuple):
	... 	x: float
	... 	y: float
	>>> x, y = Var[float, Ctx]("x"), Var[float, Ctx]("y")
	>>> def my_sqrt(v: float) -> float:
	... 	if v < 0: raise ValueError(f"neg: {v}")
	... 	return v ** 0.5
	>>> safe_sqrt = python_func(my_sqrt)

	Success:

	>>> safe_sqrt(x).to_string(Ctx(x=4.0, y=0.0))
	'my_sqrt(x:4.0) -> 2.0'

	Exception caught as Failure:

	>>> safe_sqrt(x).to_string(Ctx(x=-1.0, y=0.0))
	"my_sqrt(x:-1.0) -> Failure(exceptions=(ValueError('neg: -1.0'),))"

	Nested calls accumulate failures from all arguments:

	>>> def div(a: float, b: float) -> float:
	... 	return a / b
	>>> safe_div = python_func(div)
	>>> r = safe_div(safe_sqrt(x), safe_sqrt(y)).unwrap(Ctx(x=-1.0, y=-4.0))
	>>> isinstance(r, Failure) and len(r.exceptions)
	2
	"""

	func: Callable[..., R]
	args: tuple[Expr[Any, S, Any], ...]

	@property
	def name(self) -> str:
		return _func_name(self.func)

	def eval(self, ctx: S) -> Const[R]:  # pyright: ignore[reportIncompatibleMethodOverride]
		vals = tuple(arg.unwrap(ctx) for arg in self.args)
		failures = tuple(e for v in vals if isinstance(v, Failure) for e in v.exceptions)
		if failures:
			return Failure(failures)  # type: ignore[return-value]
		try:
			return Const(None, self.func(*vals))
		except Exception as e:
			return Failure((e,))  # type: ignore[return-value]

	def to_string(self, ctx: S | None = None) -> str:
		args_str = ", ".join(a.to_string(ctx) for a in self.args)
		if ctx is None:
			return f"{self.name}({args_str})"
		return f"{self.name}({args_str}) -> {self.unwrap(ctx)}"

	def partial(self, ctx: Any) -> "PythonFuncBase[R, Any]":  # pyright: ignore[reportIncompatibleMethodOverride]
		return type(self)(self.func, tuple(arg.partial(ctx) for arg in self.args))  # type: ignore[call-arg]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc0[R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[], R]
	args: tuple[Expr[Any, S, Any], ...] = ()


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc1[T1, R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[T1], R]
	args: tuple[Expr[Any, S, T1] | Expr[Any, S, T1 | Failure]]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc2[T1, T2, R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[T1, T2], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc3[T1, T2, T3, R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[T1, T2, T3], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc4[T1, T2, T3, T4, R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[T1, T2, T3, T4], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc5[T1, T2, T3, T4, T5, R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[T1, T2, T3, T4, T5], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc6[T1, T2, T3, T4, T5, T6, R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[T1, T2, T3, T4, T5, T6], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc7[T1, T2, T3, T4, T5, T6, T7, R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[T1, T2, T3, T4, T5, T6, T7], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc8[T1, T2, T3, T4, T5, T6, T7, T8, R, S: ContextProtocol](PythonFuncBase[R, S]):
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc9[T1, T2, T3, T4, T5, T6, T7, T8, T9, R, S: ContextProtocol](
	PythonFuncBase[R, S],
):
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
		Expr[Any, S, T9] | Expr[Any, S, T9 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc10[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, R, S: ContextProtocol](
	PythonFuncBase[R, S],
):
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
		Expr[Any, S, T9] | Expr[Any, S, T9 | Failure],
		Expr[Any, S, T10] | Expr[Any, S, T10 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc11[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, R, S: ContextProtocol](
	PythonFuncBase[R, S],
):
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
		Expr[Any, S, T9] | Expr[Any, S, T9 | Failure],
		Expr[Any, S, T10] | Expr[Any, S, T10 | Failure],
		Expr[Any, S, T11] | Expr[Any, S, T11 | Failure],
	]


@dataclass(frozen=True, eq=False, slots=True)
class PythonFunc12[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, R, S: ContextProtocol](
	PythonFuncBase[R, S],
):
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12], R]
	args: tuple[
		Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
		Expr[Any, S, T9] | Expr[Any, S, T9 | Failure],
		Expr[Any, S, T10] | Expr[Any, S, T10 | Failure],
		Expr[Any, S, T11] | Expr[Any, S, T11 | Failure],
		Expr[Any, S, T12] | Expr[Any, S, T12 | Failure],
	]


@dataclass(frozen=True, slots=True)
class PythonFunc0Wrapper[R]:
	func: Callable[[], R]

	def __call__[S: ContextProtocol](self) -> PythonFunc0[R, S]:  # pyright: ignore[reportInvalidTypeVarUse]
		return PythonFunc0(self.func)


@dataclass(frozen=True, slots=True)
class PythonFunc1Wrapper[T1, R]:
	func: Callable[[T1], R]

	def __call__[S: ContextProtocol](
		self, arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure]
	) -> PythonFunc1[T1, R, S]:
		return PythonFunc1(self.func, (arg1 if isinstance(arg1, Expr) else Const(None, arg1),))  # pyright: ignore[reportUnknownArgumentType]


@dataclass(frozen=True, slots=True)
class PythonFunc2Wrapper[T1, T2, R]:
	func: Callable[[T1, T2], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
	) -> PythonFunc2[T1, T2, R, S]:
		return PythonFunc2(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc3Wrapper[T1, T2, T3, R]:
	func: Callable[[T1, T2, T3], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
	) -> PythonFunc3[T1, T2, T3, R, S]:
		return PythonFunc3(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc4Wrapper[T1, T2, T3, T4, R]:
	func: Callable[[T1, T2, T3, T4], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
	) -> PythonFunc4[T1, T2, T3, T4, R, S]:
		return PythonFunc4(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc5Wrapper[T1, T2, T3, T4, T5, R]:
	func: Callable[[T1, T2, T3, T4, T5], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		arg5: T5 | Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
	) -> PythonFunc5[T1, T2, T3, T4, T5, R, S]:
		return PythonFunc5(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
				arg5 if isinstance(arg5, Expr) else Const(None, arg5),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc6Wrapper[T1, T2, T3, T4, T5, T6, R]:
	func: Callable[[T1, T2, T3, T4, T5, T6], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		arg5: T5 | Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		arg6: T6 | Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
	) -> PythonFunc6[T1, T2, T3, T4, T5, T6, R, S]:
		return PythonFunc6(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
				arg5 if isinstance(arg5, Expr) else Const(None, arg5),  # pyright: ignore[reportUnknownArgumentType]
				arg6 if isinstance(arg6, Expr) else Const(None, arg6),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc7Wrapper[T1, T2, T3, T4, T5, T6, T7, R]:
	func: Callable[[T1, T2, T3, T4, T5, T6, T7], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		arg5: T5 | Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		arg6: T6 | Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		arg7: T7 | Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
	) -> PythonFunc7[T1, T2, T3, T4, T5, T6, T7, R, S]:
		return PythonFunc7(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
				arg5 if isinstance(arg5, Expr) else Const(None, arg5),  # pyright: ignore[reportUnknownArgumentType]
				arg6 if isinstance(arg6, Expr) else Const(None, arg6),  # pyright: ignore[reportUnknownArgumentType]
				arg7 if isinstance(arg7, Expr) else Const(None, arg7),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc8Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, R]:
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		arg5: T5 | Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		arg6: T6 | Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		arg7: T7 | Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		arg8: T8 | Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
	) -> PythonFunc8[T1, T2, T3, T4, T5, T6, T7, T8, R, S]:
		return PythonFunc8(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
				arg5 if isinstance(arg5, Expr) else Const(None, arg5),  # pyright: ignore[reportUnknownArgumentType]
				arg6 if isinstance(arg6, Expr) else Const(None, arg6),  # pyright: ignore[reportUnknownArgumentType]
				arg7 if isinstance(arg7, Expr) else Const(None, arg7),  # pyright: ignore[reportUnknownArgumentType]
				arg8 if isinstance(arg8, Expr) else Const(None, arg8),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc9Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, T9, R]:
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		arg5: T5 | Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		arg6: T6 | Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		arg7: T7 | Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		arg8: T8 | Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
		arg9: T9 | Expr[Any, S, T9] | Expr[Any, S, T9 | Failure],
	) -> PythonFunc9[T1, T2, T3, T4, T5, T6, T7, T8, T9, R, S]:
		return PythonFunc9(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
				arg5 if isinstance(arg5, Expr) else Const(None, arg5),  # pyright: ignore[reportUnknownArgumentType]
				arg6 if isinstance(arg6, Expr) else Const(None, arg6),  # pyright: ignore[reportUnknownArgumentType]
				arg7 if isinstance(arg7, Expr) else Const(None, arg7),  # pyright: ignore[reportUnknownArgumentType]
				arg8 if isinstance(arg8, Expr) else Const(None, arg8),  # pyright: ignore[reportUnknownArgumentType]
				arg9 if isinstance(arg9, Expr) else Const(None, arg9),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc10Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, R]:
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		arg5: T5 | Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		arg6: T6 | Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		arg7: T7 | Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		arg8: T8 | Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
		arg9: T9 | Expr[Any, S, T9] | Expr[Any, S, T9 | Failure],
		arg10: T10 | Expr[Any, S, T10] | Expr[Any, S, T10 | Failure],
	) -> PythonFunc10[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, R, S]:
		return PythonFunc10(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
				arg5 if isinstance(arg5, Expr) else Const(None, arg5),  # pyright: ignore[reportUnknownArgumentType]
				arg6 if isinstance(arg6, Expr) else Const(None, arg6),  # pyright: ignore[reportUnknownArgumentType]
				arg7 if isinstance(arg7, Expr) else Const(None, arg7),  # pyright: ignore[reportUnknownArgumentType]
				arg8 if isinstance(arg8, Expr) else Const(None, arg8),  # pyright: ignore[reportUnknownArgumentType]
				arg9 if isinstance(arg9, Expr) else Const(None, arg9),  # pyright: ignore[reportUnknownArgumentType]
				arg10 if isinstance(arg10, Expr) else Const(None, arg10),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc11Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, R]:
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		arg5: T5 | Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		arg6: T6 | Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		arg7: T7 | Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		arg8: T8 | Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
		arg9: T9 | Expr[Any, S, T9] | Expr[Any, S, T9 | Failure],
		arg10: T10 | Expr[Any, S, T10] | Expr[Any, S, T10 | Failure],
		arg11: T11 | Expr[Any, S, T11] | Expr[Any, S, T11 | Failure],
	) -> PythonFunc11[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, R, S]:
		return PythonFunc11(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
				arg5 if isinstance(arg5, Expr) else Const(None, arg5),  # pyright: ignore[reportUnknownArgumentType]
				arg6 if isinstance(arg6, Expr) else Const(None, arg6),  # pyright: ignore[reportUnknownArgumentType]
				arg7 if isinstance(arg7, Expr) else Const(None, arg7),  # pyright: ignore[reportUnknownArgumentType]
				arg8 if isinstance(arg8, Expr) else Const(None, arg8),  # pyright: ignore[reportUnknownArgumentType]
				arg9 if isinstance(arg9, Expr) else Const(None, arg9),  # pyright: ignore[reportUnknownArgumentType]
				arg10 if isinstance(arg10, Expr) else Const(None, arg10),  # pyright: ignore[reportUnknownArgumentType]
				arg11 if isinstance(arg11, Expr) else Const(None, arg11),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@dataclass(frozen=True, slots=True)
class PythonFunc12Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, R]:
	func: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12], R]

	def __call__[S: ContextProtocol](
		self,
		arg1: T1 | Expr[Any, S, T1] | Expr[Any, S, T1 | Failure],
		arg2: T2 | Expr[Any, S, T2] | Expr[Any, S, T2 | Failure],
		arg3: T3 | Expr[Any, S, T3] | Expr[Any, S, T3 | Failure],
		arg4: T4 | Expr[Any, S, T4] | Expr[Any, S, T4 | Failure],
		arg5: T5 | Expr[Any, S, T5] | Expr[Any, S, T5 | Failure],
		arg6: T6 | Expr[Any, S, T6] | Expr[Any, S, T6 | Failure],
		arg7: T7 | Expr[Any, S, T7] | Expr[Any, S, T7 | Failure],
		arg8: T8 | Expr[Any, S, T8] | Expr[Any, S, T8 | Failure],
		arg9: T9 | Expr[Any, S, T9] | Expr[Any, S, T9 | Failure],
		arg10: T10 | Expr[Any, S, T10] | Expr[Any, S, T10 | Failure],
		arg11: T11 | Expr[Any, S, T11] | Expr[Any, S, T11 | Failure],
		arg12: T12 | Expr[Any, S, T12] | Expr[Any, S, T12 | Failure],
	) -> PythonFunc12[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, R, S]:
		return PythonFunc12(
			self.func,
			(
				arg1 if isinstance(arg1, Expr) else Const(None, arg1),  # pyright: ignore[reportUnknownArgumentType]
				arg2 if isinstance(arg2, Expr) else Const(None, arg2),  # pyright: ignore[reportUnknownArgumentType]
				arg3 if isinstance(arg3, Expr) else Const(None, arg3),  # pyright: ignore[reportUnknownArgumentType]
				arg4 if isinstance(arg4, Expr) else Const(None, arg4),  # pyright: ignore[reportUnknownArgumentType]
				arg5 if isinstance(arg5, Expr) else Const(None, arg5),  # pyright: ignore[reportUnknownArgumentType]
				arg6 if isinstance(arg6, Expr) else Const(None, arg6),  # pyright: ignore[reportUnknownArgumentType]
				arg7 if isinstance(arg7, Expr) else Const(None, arg7),  # pyright: ignore[reportUnknownArgumentType]
				arg8 if isinstance(arg8, Expr) else Const(None, arg8),  # pyright: ignore[reportUnknownArgumentType]
				arg9 if isinstance(arg9, Expr) else Const(None, arg9),  # pyright: ignore[reportUnknownArgumentType]
				arg10 if isinstance(arg10, Expr) else Const(None, arg10),  # pyright: ignore[reportUnknownArgumentType]
				arg11 if isinstance(arg11, Expr) else Const(None, arg11),  # pyright: ignore[reportUnknownArgumentType]
				arg12 if isinstance(arg12, Expr) else Const(None, arg12),  # pyright: ignore[reportUnknownArgumentType]
			),
		)


@overload
def python_func[R](f: Callable[[], R]) -> PythonFunc0Wrapper[R]: ...


@overload
def python_func[T1, R](f: Callable[[T1], R]) -> PythonFunc1Wrapper[T1, R]: ...


@overload
def python_func[T1, T2, R](f: Callable[[T1, T2], R]) -> PythonFunc2Wrapper[T1, T2, R]: ...


@overload
def python_func[T1, T2, T3, R](
	f: Callable[[T1, T2, T3], R],
) -> PythonFunc3Wrapper[T1, T2, T3, R]: ...


@overload
def python_func[T1, T2, T3, T4, R](
	f: Callable[[T1, T2, T3, T4], R],
) -> PythonFunc4Wrapper[T1, T2, T3, T4, R]: ...


@overload
def python_func[T1, T2, T3, T4, T5, R](
	f: Callable[[T1, T2, T3, T4, T5], R],
) -> PythonFunc5Wrapper[T1, T2, T3, T4, T5, R]: ...


@overload
def python_func[T1, T2, T3, T4, T5, T6, R](
	f: Callable[[T1, T2, T3, T4, T5, T6], R],
) -> PythonFunc6Wrapper[T1, T2, T3, T4, T5, T6, R]: ...


@overload
def python_func[T1, T2, T3, T4, T5, T6, T7, R](
	f: Callable[[T1, T2, T3, T4, T5, T6, T7], R],
) -> PythonFunc7Wrapper[T1, T2, T3, T4, T5, T6, T7, R]: ...


@overload
def python_func[T1, T2, T3, T4, T5, T6, T7, T8, R](
	f: Callable[[T1, T2, T3, T4, T5, T6, T7, T8], R],
) -> PythonFunc8Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, R]: ...


@overload
def python_func[T1, T2, T3, T4, T5, T6, T7, T8, T9, R](
	f: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9], R],
) -> PythonFunc9Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, T9, R]: ...


@overload
def python_func[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, R](
	f: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10], R],
) -> PythonFunc10Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, R]: ...


@overload
def python_func[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, R](
	f: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11], R],
) -> PythonFunc11Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, R]: ...


@overload
def python_func[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, R](
	f: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12], R],
) -> PythonFunc12Wrapper[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, R]: ...


def python_func(f: Callable[..., Any]) -> Any:
	"""Lift a Python callable into the Result context.

	Returns a wrapper whose ``__call__`` accepts Expr or literal arguments and
	produces a PythonFunc expression (a ResultBase).  Supports 0-12 parameters.

	>>> from typing import NamedTuple
	>>> from mahonia import Var
	>>> from mahonia.python_func import python_func
	>>> class Ctx(NamedTuple):
	... 	x: float
	>>> x = Var[float, Ctx]("x")
	>>> def my_sqrt(v: float) -> float:
	... 	if v < 0: raise ValueError(f"neg: {v}")
	... 	return v ** 0.5
	>>> safe_sqrt = python_func(my_sqrt)
	>>> safe_sqrt(x).to_string()
	'my_sqrt(x)'
	>>> safe_sqrt(x).to_string(Ctx(x=4.0))
	'my_sqrt(x:4.0) -> 2.0'
	>>> safe_sqrt(x).to_string(Ctx(x=-1.0))
	"my_sqrt(x:-1.0) -> Failure(exceptions=(ValueError('neg: -1.0'),))"
	"""
	match len(inspect.signature(f).parameters):
		case 0:
			return PythonFunc0Wrapper(f)
		case 1:
			return PythonFunc1Wrapper(f)
		case 2:
			return PythonFunc2Wrapper(f)
		case 3:
			return PythonFunc3Wrapper(f)
		case 4:
			return PythonFunc4Wrapper(f)
		case 5:
			return PythonFunc5Wrapper(f)
		case 6:
			return PythonFunc6Wrapper(f)
		case 7:
			return PythonFunc7Wrapper(f)
		case 8:
			return PythonFunc8Wrapper(f)
		case 9:
			return PythonFunc9Wrapper(f)
		case 10:
			return PythonFunc10Wrapper(f)
		case 11:
			return PythonFunc11Wrapper(f)
		case 12:
			return PythonFunc12Wrapper(f)
		case n:
			raise ValueError(f"python_func supports 0-12 args, got {n}")



import asyncio
import csv
from typing import Any, Final, NamedTuple

import cyclopts
from rich.console import Console

import mahonia
from manufacturing.types import Testable, TestStep, cached_context

app = cyclopts.App()
console = Console()


class Power(NamedTuple):
	voltage: float
	current: float


class Temperature(NamedTuple):
	temp_readings: list[float]
	temp: float


@cached_context
async def step0_check_power() -> Power:
	await asyncio.sleep(2.000)
	return Power(voltage=4.2, current=0.42)


def step1_power() -> TestStep[Power]:
	voltage = mahonia.Var[float, Power]("voltage")
	current = mahonia.Var[float, Power]("current")
	power = voltage * current

	tests = (
		(voltage > 3.0) & (voltage < 5.0),
		(current > 0.1) & (current < 1.0),
		mahonia.Predicate("Power under limit", power < 5.0),
		mahonia.Predicate("Power over minimum", power > 0.5),
	)

	return TestStep(step0_check_power().context_func, tests)


def step2_resistance() -> TestStep[Power]:
	voltage = mahonia.Var[float, Power]("voltage")
	current = mahonia.Var[float, Power]("current")
	resistance = voltage / current
	target_resistance = mahonia.PlusMinus("R_target", 10.0, 1.0)

	tests = (
		mahonia.Predicate("Resistance check", resistance == target_resistance),
		resistance > 5.0,
	)

	return TestStep(step0_check_power().context_func, tests)


def step3_temperature() -> TestStep[Temperature]:
	temps = mahonia.Var[mahonia.SizedIterable[float], Temperature]("temp_readings")
	temp = mahonia.Var[float, Temperature]("temp")
	avg_temp = mahonia.Const("avg_temp", 25.0)
	in_range = (temp > 20.0) & (temp < 30.0)
	exceeds_avg = temp > avg_temp

	tests = (
		mahonia.AllExpr(in_range.map(temps)),
		mahonia.Predicate("Some temps exceed average", mahonia.AnyExpr(exceeds_avg.map(temps))),
	)

	async def func() -> Temperature:
		await asyncio.sleep(0.5)
		return Temperature(temp_readings=[22.0, 24.0, 26.0], temp=25.0)

	return TestStep(func, tests)


STEPS: Final = (
	step0_check_power,
	step1_power,
	step2_resistance,
	step3_temperature,
)


class PassContext(NamedTuple):
	tests: mahonia.SizedIterable[mahonia.BoundExpr[Any, Any, Any]]

	@classmethod
	def from_cases(
		cls,
		tests: mahonia.SizedIterable[Testable],
		ctx: mahonia.ContextProtocol,
	) -> "PassContext":
		return cls(tests=tuple(t.bind(ctx) for t in tests))


@app.default
async def run() -> None:
	tests: Final = mahonia.Var[
		mahonia.SizedIterable[mahonia.BoundExpr[Any, Any, Any]], PassContext
	]("tests")
	pass_expr: Final = mahonia.FoldLExpr(mahonia.And, tests)

	for step in STEPS:
		console.print(f"Test step: {step.__name__}")
		for test in step().tests:
			print(f"  Test: {test.to_string()}")

	console.print()
	console.print("----------------------")
	console.print()

	class ResultContext(NamedTuple):
		results: list[mahonia.BoundExpr[Any, Any, Any]]

	results: Final = mahonia.Var[
		mahonia.SizedIterable[mahonia.BoundExpr[Any, Any, Any]], ResultContext
	]("results")

	result_context = ResultContext(results=[])

	with open("test_results.csv", "w", newline="") as csvfile:
		writer = csv.writer(csvfile)
		writer.writerow(["Step", "Pass/Fail", "Context", "Tests"])

		for step in STEPS:
			ctx = await step().context_func()
			console.print(f"Test step: {step.__name__} {ctx=}")

			if len(step().tests) == 0:
				console.print("  No tests defined, skipping.")
				continue
			pass_bound = pass_expr.bind(PassContext.from_cases(step().tests, ctx))
			result_context.results.append(pass_bound)

			pass_fail = "PASS" if pass_bound.unwrap() else "FAIL"
			writer.writerow(
				(
					step.__name__,
					pass_fail,
					str(ctx),
					str(pass_bound),
				)
			)
			for test in step().tests:
				print(f"  Test: {test.to_string(ctx)}")

	result_expr = mahonia.FoldLExpr(mahonia.And, results)

	console.print()
	console.print(f"Overall result: {'PASS' if result_expr.unwrap(result_context) else 'FAIL'}")


@app.command
async def list_tests() -> None: ...


def main() -> None:
	app()
